import psutil, random, aiohttp, socket, orjson
from typing import Optional

class NoInterface(Exception):
    def __init__(self):
        return super().__init__(f"Please provide an interface to pull addresses from using the interface= argument")

def get_local_addresses(interface:str):
    interfaces=psutil.net_if_addrs()
    return [(str(i.address),0) for i in interfaces[interface] if '.' in i.address]

def get_all_interface_addresses():
    addresses=[]
    interfaces=psutil.net_if_addrs()
    interface_keys=list(interfaces.keys())
    for interface in interface_keys:
        for i in interfaces[interface]:
            if i.address:
                if "." in i.address and not i.address.startswith('127.') and not i.address.startswith('192'):
                    addresses.append((str(i.address),0))
    return addresses

class ProxyPool(object):
    def __init__(self,proxies:Optional[list],use_local:bool=False,interface: Optional[str]=None):
        if proxies:
            self.proxies = proxies
            self.local = False
        else:
            if use_local == True:
                self.local=True
                if not interface: self.proxies = get_all_interface_addresses()
                else: self.proxies = get_local_addresses(interface)
        self.proxies=iter(self.proxies)

    async def session(self):
        if self.local == True:
            return aiohttp.ClientSession(connector=aiohttp.TCPConnector(family=socket.AF_INET,limit=0,local_addr=(next(self.proxies),0)),json_serialize=lambda x: orjson.dumps(x).decode(),resolver=aiohttp.AsyncResolver())
        else:
            return aiohttp.ClientSession(connector=aiohttp.TCPConnector(family=socket.AF_INET,limit=0),proxy=next(self.proxies),resolver=aiohttp.AsyncResolver(),json_serialize=lambda x: orjson.dumps(x).decode(),)



import asyncio,datetime,time
from typing import Any

class ExpiringDictionary:
    def __init__(self):
        self.dict={}
        self.rl={}
        self.delete={}
        self.futures={}

    async def do_expiration(self,key:str,expiration:int):
        await asyncio.sleep(expiration)
        if key in self.dict: self.dict.pop(key)

    async def do_cancel(self,key:str):
        if key in self.futures:
            try: self.futures[key].cancel()
            except: pass
            return True
        else: return False

    async def set(self,key:str,value:Any,expiration:int=60):
        if key in self.futures:
            await self.do_cancel(key)
        self.dict[key]=value
        future=asyncio.ensure_future(self.do_expiration(key,expiration))
        self.futures[key]=future
        return 1

    async def remove(self,key:str):
        if key in self.dict:
            try: await self.do_cancel(key)
            except: pass
            self.dict.pop(key)
            return 1
        else:
            return 0

    async def get(self,key:str):
        if key in self.dict:
            return self.dict[key]
        else:
            return 0

    async def keys(self):
        return list(self.dict.keys())

    async def do_delete(self,key):
       self.dict.pop(key)
       self.delete[key]['last']=int(datetime.datetime.now().timestamp())

    def is_ratelimited(self,key:str):
        if key in self.dict:
            if self.dict[key] >= self.rl[key]: return True
        return False

    def time_remaining(self,key:str):
        if key in self.dict and key in self.delete:
            if not self.dict[key] >= self.rl[key]:
                return 0
            remaining=(self.delete[key]['last']+self.delete[key]['bucket'])-int(datetime.datetime.now().timestamp())
            return remaining
        else:
            return 0

    async def ratelimit(self,key:str,amount:int,bucket:int=60):
        if key not in self.dict:
            self.dict[key]=1
            self.rl[key]=amount
            if key not in self.delete:
                self.delete[key]={'bucket':bucket,'last':int(datetime.datetime.now().timestamp())}
            return False
        else:
            try:
                if self.delete[key]['last'] + bucket <= int(datetime.datetime.now().timestamp()):
                    self.dict.pop(key)
                    self.delete[key]['last']=int(datetime.datetime.now().timestamp())
                    self.dict[key]=0
                self.dict[key]+=1
                if self.dict[key] >= self.rl[key]:
                    return True
                else:
                    return False
            except:
                return await self.ratelimit(key,amount,bucket)



# THIS WAS MADE IN NANO I AM NOT PRESSING SPACE IN NANO BRAH ILL CLEAN IT UP LATER


import asyncio,datetime,humanfriendly,typing
from .member import Member
from .guild import Guild
from .user import User
import .utils as utils
from .object import Object
from xxhash import xxh3_64_hexdigest as keyhash
class InvalidType(Exception):
	def __init__(self,types:typing.Any):
		self.types = types
		super.__init___(f"Only Types Useable are {', '.join(t for t in self.types)}")

class Punishments:
	def __init__(self):
		self.thresholds = {}
		self.users = {}
		self.temp_bans = {}
		self.futures = {}


	async def add_punishment(self,guild_id:int,amount:int,punishment:str,time:int=None):
		punishments={'ban','kick','timeout'}
		if punishment not in punishments: raise InvalidType(punishments)
		if guild_id not in self.thresholds: self.thresholds[guild_id]={}
		if guild_id not in self.users: self.users[guild_id]={}
		if punishment != "timeout": self.thresholds[guild_id][amount]=punishment
		else: self.thresholds[guild_id][amount]=f"timeout-{time}"
		return True

	async def remove_punishment(self,guild_id:int,amount:int):
		if guild_id not in self.thresholds: return False
		if amount not in self.thresholds[guild_id]: return False
		self.thresholds[guild_id].pop(amount)
		return True

	async def clear_threshold(self,guild:Guild,member:typing.Union[Member,User]):
		if guild.id not in self.users: return 0
		if member.id not in self.users[guild.id]: return 0
		self.users[guild.id][member.id]=0
		return 0

	async def do_unban(self,guild:Guild,time:int,user:int):
		await asyncio.sleep(time)
		return await guild.smart_unban(user)

	async def cancel_tempban(self,guild:Guild,member:User):
		key=keyhash(f"tb{guild.id}{member.id}")
		if key in self.futures: self.futures[key].cancel()
		else: return False
		return True

	async def tempban(self,guild:Guild,member:Member,time:str,*,reason:str=None):
		now=int(datetime.datetime.now().timestamp())
		if guild.id not in self.temp_bans: self.temp_bans[gulld.id={}
		convert=humanfriendly.parse_timespan(time)
		until=datetime.datetime.now()+datetime.timedelta(seconds=convert)
		self.temp_bans[guild.id][member.id]=int(until.timestamp())
		await member.ban(reason=reason)
		t=int(until.timestamp())-now
		future=asyncio.ensure_future(self.do_unban(guild,int(t),member.id)
		hash=keyhash(f"tb{guild.id}{member.id}")
		if hash in self.futures: self.futures[hash].cancel()
		self.futures[hash]=future
		return True

	async def decide_punishment(self,guild:Guild,amount:int):
		ret=None
		for s in sorted(self.thresholds[guild.id].keys(),reverse=True):
			if amount >= key:
				ret = key
				break
		if ret: return self.thresholds[guild.id][ret]
		return ret

	async def punish(self,guild:Guild,member:Member,*,reason:str=None):
		if guild.id in self.thresholds:
			if guild.id not in self.users: self.users[guild.id]={}
			if member.id in self.users[guild_id]
				self.users[guild.id][member.id]+=1
			else: self.users[guild.id][member.id]=1
			amounts=list(self.thresholds[guild.id].keys())
			current=self.users[guild.id][member.id]
			punishment=await self.decide_punishment(guild,current)
			if punishment:
				if punishment == "ban":
					return await guild.ban(Object(member.id),reason=reason)
				if punishment == "kick":
					return await member.kick(reason=reason)
				else:
					time=int(punishment.split("-")[1])
					return await member.timeout(utils.utcnow()+datetime.timedelta(seconds=time),reason=reason)
			else:
				return False
